package ca.carleton.gcrc.couch.onUpload;

import java.io.File;
import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Logger;

import net.sf.json.JSONObject;

import ca.carleton.gcrc.couch.client.CouchDesignDocument;
import ca.carleton.gcrc.couch.client.CouchQuery;
import ca.carleton.gcrc.couch.client.CouchQueryResults;
import ca.carleton.gcrc.couch.onUpload.mail.MailNotification;
import ca.carleton.gcrc.olkit.multimedia.converter.MultimediaConversionRequest;
import ca.carleton.gcrc.olkit.multimedia.converter.MultimediaConverter;
import ca.carleton.gcrc.olkit.multimedia.converter.impl.MultimediaConverterImpl;
import ca.carleton.gcrc.olkit.multimedia.file.SystemFile;
import ca.carleton.gcrc.olkit.multimedia.utils.MimeUtils;
import ca.carleton.gcrc.olkit.multimedia.utils.MimeUtils.MultimediaClass;

public class UploadWorkerThread extends Thread {

	final protected Logger logger = Logger.getLogger(this.getClass());
	
	private boolean isShuttingDown = false;
	private CouchDesignDocument dd;
	private File mediaDir;
	private MailNotification mailNotification;
	private Set<String> docIdsToSkip = new HashSet<String>();
	private MultimediaConverter mmConverter = new MultimediaConverterImpl();
	
	protected UploadWorkerThread(
		CouchDesignDocument dd
		,File mediaDir
		,MailNotification mailNotification
		) {
		this.dd = dd;
		this.mediaDir = mediaDir;
		this.mailNotification = mailNotification;
	}
	
	public void shutdown() {
		
		logger.info("Shutting down upload worker thread");

		synchronized(this) {
			isShuttingDown = true;
			this.notifyAll();
		}
	}
	
	@Override
	public void run() {
		
		logger.info("Start upload worker thread");
		
		boolean done = false;
		do {
			synchronized(this) {
				done = isShuttingDown;
			}
			if( false == done ) {
				activity();
			}
		} while( false == done );

		logger.info("Upload worker thread exiting");
	}
	
	private void activity() {
		CouchQuery query = new CouchQuery();
		query.setViewName("server_work");
		
		CouchQueryResults results;
		try {
			results = dd.performQuery(query);
		} catch (Exception e) {
			logger.error("Error accessing server",e);
			waitMillis(60 * 1000); // wait a minute
			return;
		}
		
		// Check for work
		String docId = null;
		String state = null;
		for(JSONObject row : results.getRows()) {
			String id = row.optString("id");
			if( false == docIdsToSkip.contains(id) ) {
				// Found some work
				docId = id;
				state = row.optString("key");
				break;
			}
		}
		
		if( null == docId ) {
			// Nothing to do, wait 5 secs
			waitMillis(5 * 1000);
			return;
		} else {
			try {
				// Handle this work
				performWork(docId, state);
				
			} catch(Exception e) {
				logger.error("Error processing document "+docId+" ("+state+")",e);
				docIdsToSkip.add(docId);
			}
		}
	}
	
	private void performWork(String docId, String state) throws Exception {
		
		logger.info("Upload worker thread processing: "+docId+" ("+state+")");
		
		if( UploadConstants.UPLOAD_VALUE_SUBMITTED.equals(state) ) {
			performSubmittedWork(docId);
			
		} else if( UploadConstants.UPLOAD_VALUE_ANALYZED.equals(state) ) {
			performAnalyzedWork(docId);
			
		} else if( UploadConstants.UPLOAD_VALUE_APPROVED.equals(state) ) {
			performApprovedWork(docId);
				
		} else {
			throw new Exception("Unrecognized state: "+state);
		}

		logger.info("Upload worker thread completed: "+docId+" ("+state+")");
	}

	private void performSubmittedWork(String docId) throws Exception {
		JSONObject doc = dd.getDatabase().getDocument(docId);
		JSONObject originalObj = doc.getJSONObject("original");
		
		// Figure out file
		File file = new File(mediaDir, originalObj.getString("media"));
		
		long fileSize = file.length();		
		SystemFile sf = SystemFile.getSystemFile(file);
		
		doc.put(UploadConstants.UPLOAD_KEY, UploadConstants.UPLOAD_VALUE_ANALYZED);
		originalObj.put(UploadConstants.SIZE_KEY, fileSize);
		originalObj.put(UploadConstants.MIME_KEY, sf.getMimeType());
		originalObj.put(UploadConstants.ENCODING_KEY, sf.getMimeEncoding());
		
		// Is it a known MIME type?
		{
			MultimediaClass aClass = MimeUtils.getMultimediaClassFromMimeType(sf.getMimeType());
			if( MultimediaClass.AUDIO == aClass 
			 || MultimediaClass.VIDEO == aClass 
			 || MultimediaClass.IMAGE == aClass 
			 ) {
				doc.put(UploadConstants.MIME_CLASS_KEY, aClass.getValue());
			}
		}
		
		dd.getDatabase().updateDocument(doc);
	}

	private void performAnalyzedWork(String docId) throws Exception {
		JSONObject doc = dd.getDatabase().getDocument(docId);
		JSONObject originalObj = doc.getJSONObject("original");
		
		// Figure out file
		File file = new File(mediaDir, originalObj.getString("media"));
		String mimeType = originalObj.getString(UploadConstants.MIME_KEY);

		// Perform conversion(s)
		MultimediaConversionRequest request = new MultimediaConversionRequest();
		request.setInFile( file );
		request.setThumbnailRequested(true);
		request.setProgress( new MultimediaProgressAdaptor() );
		
		MultimediaClass mmClass = MimeUtils.getMultimediaClassFromMimeType(mimeType);
		if( MultimediaClass.VIDEO == mmClass ) {
			// For video file, convert to appropriate file type
			mmConverter.convertVideo(request);

		} else if( MultimediaClass.AUDIO == mmClass ) {
			// For audio file, convert to appropriate file type
			mmConverter.convertAudio(request);
			
		} else if( MultimediaClass.IMAGE == mmClass ) {
			// For image file, convert to appropriate file type
			mmConverter.convertImage(request);
			
		} else {
			// No conversion required. Go directly to waiting for approval
			doc.put(UploadConstants.UPLOAD_KEY, UploadConstants.UPLOAD_VALUE_WAITING_FOR_APPROVAL);
			dd.getDatabase().updateDocument(doc);
			sendContributionApprovalRequest(docId, doc);
			return;
		}
		
		// Report original size
		if( request.getInHeight() != 0 && request.getInWidth() != 0 ) {
			originalObj.put("height", request.getInHeight());
			originalObj.put("width", request.getInWidth());
		}
		

		// Report converted object
		{
			File convertedFile = request.getOutFile();
			SystemFile convertedSf = SystemFile.getSystemFile(convertedFile);

			JSONObject convertedObj = new JSONObject();
			convertedObj.put("conversionPerformed", request.isConversionPerformed());
			convertedObj.put("media", convertedFile.getName());
			convertedObj.put(UploadConstants.SIZE_KEY, convertedFile.length());
			convertedObj.put(UploadConstants.MIME_KEY, convertedSf.getMimeType());
			convertedObj.put(UploadConstants.ENCODING_KEY, convertedSf.getMimeEncoding());

			if( request.getOutHeight() != 0 && request.getOutWidth() != 0 ) {
				convertedObj.put("height", request.getOutHeight());
				convertedObj.put("width", request.getOutWidth());
			}
			
			doc.put("converted",convertedObj);
		}

		// Report thumbnail object
		if( request.isThumbnailCreated() ) {
			File thumbFile = request.getThumbnailFile();
			SystemFile thumbSf = SystemFile.getSystemFile(thumbFile);

			// Report converted object
			JSONObject thumbObj = new JSONObject();
			thumbObj.put("media", thumbFile.getName());
			thumbObj.put(UploadConstants.SIZE_KEY, thumbFile.length());
			thumbObj.put(UploadConstants.MIME_KEY, thumbSf.getMimeType());
			thumbObj.put(UploadConstants.ENCODING_KEY, thumbSf.getMimeEncoding());

			if( request.getThumbnailHeight() != 0 && request.getThumbnailWidth() != 0 ) {
				thumbObj.put("height", request.getThumbnailHeight());
				thumbObj.put("width", request.getThumbnailWidth());
			}

			doc.put("thumbnail",thumbObj);
		}
		
		// Mark end of this phase
		doc.put(UploadConstants.UPLOAD_KEY, UploadConstants.UPLOAD_VALUE_WAITING_FOR_APPROVAL);
		doc.put(UploadConstants.MIME_CLASS_KEY, mmClass.getValue());

		dd.getDatabase().updateDocument(doc);
		
		// Notify that upload is available
		sendContributionApprovalRequest(docId, doc);
	}

	private void performApprovedWork(String docId) throws Exception {
		JSONObject doc = dd.getDatabase().getDocument(docId);
		
		// Attach converted file
		if( doc.containsKey("converted") ) {
			JSONObject convertedObj = doc.getJSONObject("converted");

			File file = new File(mediaDir, convertedObj.getString("media"));
			String contentType = convertedObj.getString("mimeType");
			dd.getDatabase().uploadAttachment(doc, "converted", file, contentType);
			
			// Refresh doc
			doc = dd.getDatabase().getDocument(docId);
			
		} else if( doc.containsKey("original") ) {
			// If no conversion, then attach original file
			JSONObject originalObj = doc.getJSONObject("original");
			
			// Figure out file
			File file = new File(mediaDir, originalObj.getString("media"));
			String contentType = originalObj.getString("mimeType");
			dd.getDatabase().uploadAttachment(doc, "original", file, contentType);
			
			// Refresh doc
			doc = dd.getDatabase().getDocument(docId);

		} else {
			throw new Exception("Do not know how to handle approved file");
		}

		// Attach thumbnail
		if( doc.containsKey("thumbnail") ) {
			JSONObject thumbnailObj = doc.getJSONObject("thumbnail");

			File file = new File(mediaDir, thumbnailObj.getString("media"));
			String contentType = thumbnailObj.getString("mimeType");
			dd.getDatabase().uploadAttachment(doc, "thumbnail", file, contentType);
			
			// Refresh doc
			doc = dd.getDatabase().getDocument(docId);
		}
		
		// Mark end of this phase
		doc.put(UploadConstants.UPLOAD_KEY, UploadConstants.UPLOAD_VALUE_ATTACHED);

		dd.getDatabase().updateDocument(doc);
	}
	
	private void sendContributionApprovalRequest(String docId, JSONObject doc) {
		// Notify that upload is available
		try {
			String title = "[not set]";
			String description = "[not set]";
			
			JSONObject contributionObj = doc.optJSONObject("contribution");
			if( null != contributionObj ) {
				if( contributionObj.has("title") ) {
					title = contributionObj.getString("title");
				}
				if( contributionObj.has("description") ) {
					description = contributionObj.getString("description");
				}
			}
			
			mailNotification.uploadNotification(docId, title, description);
			
		} catch(Exception e) {
			logger.error("Failed mail notification",e);
		}
	}

	private boolean waitMillis(int millis) {
		synchronized(this) {
			if( true == isShuttingDown ) {
				return false;
			}
			
			try {
				this.wait(millis);
			} catch (InterruptedException e) {
				// Interrupted
				return false;
			}
		}
		
		return true;
	}
}
